#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from lxml import etree
import os
from requests import request

from nova import context as nova_context
from nova import exception
from nova.objects import instance
from nova.objects import share_mapping
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional.api import client
from nova.tests.functional.libvirt import base

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class ServerSharesTestBase(base.ServersTestBase):
    api_major_version = 'v2.1'
    microversion = 'latest'
    ADMIN_API = True
    FAKE_LIBVIRT_VERSION = 8000000
    FAKE_QEMU_VERSION = 6002000

    def setUp(self):
        super(ServerSharesTestBase, self).setUp()

        self.context = nova_context.get_admin_context()
        self.manila_fixture = self.useFixture(nova_fixtures.ManilaFixture())
        self.flags(ram_allocation_ratio=1.0)
        self.flags(file_backed_memory=8192, group='libvirt')
        self.compute = self.start_compute(
            'host1',
            libvirt_version=self.FAKE_LIBVIRT_VERSION,
            qemu_version=self.FAKE_QEMU_VERSION
        )

        self.api_fixture = self.useFixture(nova_fixtures.OSMetadataServer())
        self.md_url = self.api_fixture.md_url

        self.host = self.computes[self.compute].driver._host

        self.mock_connect = self.useFixture(fixtures.MockPatch(
            'nova.virt.libvirt.volume.nfs.LibvirtNFSVolumeDriver.'
            'connect_volume')).mock

        self.mock_disconnect = self.useFixture(fixtures.MockPatch(
            'nova.virt.libvirt.volume.nfs.LibvirtNFSVolumeDriver.'
            'disconnect_volume')).mock

        self.mock_connect_ceph = self.useFixture(fixtures.MockPatch(
            'nova.virt.libvirt.volume.cephfs.LibvirtCEPHFSVolumeDriver'
            '.connect_volume')).mock

        self.mock_disconnect_ceph = self.useFixture(fixtures.MockPatch(
            'nova.virt.libvirt.volume.cephfs.LibvirtCEPHFSVolumeDriver'
            '.disconnect_volume')).mock

    def _get_xml(self, server):
        self.instance = instance.Instance.get_by_uuid(
            self.context, server['id'])
        guest = self.host.get_guest(self.instance)
        xml = guest.get_xml_desc()
        return xml

    def _get_filesystem_tag(self, xml, tag):
        tags = []
        tree = etree.fromstring(xml)
        device_nodes = tree.find('./devices')
        filesystems = device_nodes.findall('./filesystem')
        for filesystem in filesystems:
            target = filesystem.find('./target')
            tags.append(target.get('dir'))
        return tags

    def _assert_filesystem_tag(self, xml, tag):
        tags = self._get_filesystem_tag(xml, tag)
        self.assertIn(tag, tags)

    def _assert_filesystem_tag_not_present(self, xml, tag):
        tags = self._get_filesystem_tag(xml, tag)
        self.assertNotIn(tag, tags)

    def _get_metadata_url(self, server):
        # make sure that the metadata service returns information about the
        # server we created above

        def fake_get_fixed_ip_by_address(self, ctxt, address):
            return {'instance_uuid': server['id']}

        self.useFixture(
            fixtures.MonkeyPatch(
                'nova.network.neutron.API.get_fixed_ip_by_address',
                fake_get_fixed_ip_by_address))
        url = '%sopenstack/latest/meta_data.json' % self.md_url
        return url

    def _assert_share_in_metadata(self, metatdata_url, share_id, tag):
        device_share_and_tag = []
        res = request('GET', metatdata_url, timeout=5)
        self.assertEqual(200, res.status_code)
        metadata = jsonutils.loads(res.text)
        for device in metadata['devices']:
            device_share_and_tag.append((device['share_id'], device['tag']))
        self.assertIn((share_id, tag), device_share_and_tag)

    def _check_traits(self):
        expected_traits = {
            "COMPUTE_STORAGE_VIRTIO_FS", "COMPUTE_MEM_BACKING_FILE"
        }

        traits = self._get_provider_traits(self.compute_rp_uuids[self.compute])

        self.assertEqual(expected_traits, expected_traits.intersection(traits))


class ServerSharesTest(ServerSharesTestBase):

    def test_server_share_metadata(self):
        """Verify that share metadata are available"""
        self._check_traits()
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'
        self._attach_share(server, share_id)
        self._start_server(server)

        # tag is the filesystem target directory.
        # if post /server/{server_id}/share was called without a specific
        # tag then the tag is the share id.
        self._assert_filesystem_tag(self._get_xml(server), share_id)

        self._assert_share_in_metadata(
            self._get_metadata_url(server), share_id, share_id)
        return (server, share_id)

    def test_server_share_metadata_with_tag(self):
        """Verify that share metadata are available with the provided tag"""
        self._check_traits()
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'
        tag = 'mytag'
        self._attach_share(server, share_id, tag=tag)
        self._start_server(server)

        # tag is the filesystem target directory.
        # if post /server/{server_id}/share was called without a specific
        # tag then the tag is the share id.
        self._assert_filesystem_tag(self._get_xml(server), tag)

        self._assert_share_in_metadata(
            self._get_metadata_url(server), share_id, tag)
        return (server, share_id)

    def test_server_share_fails_with_tag_already_used(self):
        """Verify that share create fails if we use an already assigned tag"""
        self._check_traits()
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'
        tag = 'mytag'
        self._attach_share(server, share_id, tag=tag)

        share_id = '1457bd85-7e7f-4835-92de-47834e1516b5'
        tag = 'mytag'

        exc = self.assertRaises(
            client.OpenStackApiException,
            self._attach_share,
            server,
            share_id,
            tag=tag
        )

        self.assertEqual(409, exc.response.status_code)
        self.assertIn(
            "Share '1457bd85-7e7f-4835-92de-47834e1516b5' or "
            "tag 'mytag' already associated to this server.",
            str(exc.response.text),
        )

    def test_server_cephfs_share_metadata(self):
        """Verify that cephfs share metadata are available"""
        # update the mock to call the cephfs fake values
        self.manila_fixture.mock_get.side_effect = (
            self.manila_fixture.fake_get_cephfs
        )
        self.manila_fixture.mock_get_access.side_effect = (
            self.manila_fixture.fake_get_access_cephfs
        )

        self._check_traits()
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'
        self._attach_share(server, share_id)
        self._start_server(server)

        # tag is the filesystem target directory.
        # if post /server/{server_id}/share was called without a specific
        # tag then the tag is the share id.
        self._assert_filesystem_tag(self._get_xml(server), share_id)

        self._assert_share_in_metadata(
            self._get_metadata_url(server), share_id, share_id)
        return (server, share_id)

    def test_server_share_after_hard_reboot(self):
        """Verify that share is still available after a reboot"""
        server, share_id = self.test_server_share_metadata()
        self._reboot_server(server, hard=True)

        self._assert_filesystem_tag(self._get_xml(server), share_id)

        self._assert_share_in_metadata(
            self._get_metadata_url(server), share_id, share_id)

    def test_server_share_mount_failure(self):
        os.environ['OS_DEBUG'] = "true"
        self._check_traits()

        self.mock_connect.side_effect = processutils.ProcessExecutionError
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'

        self._attach_share(server, share_id)
        response = self._get_share(server, share_id)

        self.assertEqual(
            response["share_id"], "4b021746-d0eb-4031-92aa-23c3bec182cd"
        )
        self.assertEqual(response["status"], "inactive")

        # Here we are using CastAsCallFixture so we got an exception from
        # nova compute. This should not happen without the fixture and
        # the api should answer with a 201 status code.
        exc = self.assertRaises(
            client.OpenStackApiException,
            self._start_server,
            server
        )

        self.assertIn("nova.exception.ShareMountError", str(exc))

        log_out = self.stdlog.logger.output

        self.assertIn(
            "Share id 4b021746-d0eb-4031-92aa-23c3bec182cd mount error "
            "from server",
            log_out)

        sm = share_mapping.ShareMapping.get_by_instance_uuid_and_share_id(
            self.context, server['id'], share_id)
        self.assertEqual(sm.status, 'error')
        self.instance = instance.Instance.get_by_uuid(
            self.context, server['id'])
        self.assertEqual(self.instance.vm_state, 'error')
        return (server, share_id)

    def test_server_start_fails_share_in_error(self):
        """Ensure a server can not start if its attached share is in error
           status and hard reboot attempts allow to restart the server as
           soon as the share issue is fixed.
        """
        server, share_id = self.test_server_share_mount_failure()
        self._verify_start_fails_share_in_error(server, share_id)

        server, share_id = self.test_server_share_umount_failure()
        self._verify_start_fails_share_in_error(server, share_id)

    def _verify_start_fails_share_in_error(self, server, share_id):
        exc = self.assertRaises(
            client.OpenStackApiException,
            self._start_server,
            server
        )

        # Try to start a vm in error state.
        self.assertEqual(exc.response.status_code, 409)
        self.assertIn("Cannot \'start\' instance", exc.response.text)
        self.assertIn("while it is in vm_state error", exc.response.text)

        # Reboot to do another mount attempt and fix the error.
        # But the error is not fixed.
        self.mock_connect.side_effect = processutils.ProcessExecutionError
        # Here we are using CastAsCallFixture so we got an exception from
        # nova api. This should not happen without the fixture.
        exc = self.assertRaises(
            client.OpenStackApiException,
            self._reboot_server,
            server,
            hard=True
        )

        log_out = self.stdlog.logger.output

        self.assertIn(
            "Share id 4b021746-d0eb-4031-92aa-23c3bec182cd mount error "
            "from server",
            log_out)

        sm = share_mapping.ShareMapping.get_by_instance_uuid_and_share_id(
            self.context, server['id'], share_id)
        self.assertEqual(sm.status, 'error')
        self.instance = instance.Instance.get_by_uuid(
            self.context, server['id'])
        self.assertEqual(self.instance.vm_state, 'error')

        # Reboot to do another mount attempt and fix the error.
        # But the error is fixed now.
        self.mock_connect.side_effect = None
        self._reboot_server(server, hard=True)

        sm = share_mapping.ShareMapping.get_by_instance_uuid_and_share_id(
            self.context, server['id'], share_id)
        self.assertEqual(sm.status, 'active')
        self.instance = instance.Instance.get_by_uuid(
            self.context, server['id'])
        self.assertEqual(self.instance.vm_state, 'active')

    def test_detach_server_share_in_error(self):
        """Ensure share can still be detached even if
           the share are in an error state.
        """
        os.environ['OS_DEBUG'] = "true"
        server, share_id = self.test_server_share_umount_failure()

        # Simulate an attempt to detach that fail due to umount error.
        # In that case we should have an umount error.
        self.mock_disconnect.side_effect = processutils.ProcessExecutionError
        # Ensure we failed to umount he share
        log_out = self.stdlog.logger.output

        self.assertIn(
            "Share id 4b021746-d0eb-4031-92aa-23c3bec182cd umount error",
            log_out)

        # We detach the share. As a consequence, we are leaking the share
        # mounted on the compute.
        self._detach_share(server, share_id)

        # Share is removed so not anymore in the DB.
        self.assertRaises(
            exception.ShareNotFound,
            share_mapping.ShareMapping.get_by_instance_uuid_and_share_id,
            self.context,
            server['id'],
            share_id
        )
        self.instance = instance.Instance.get_by_uuid(
            self.context, server['id'])
        self.assertEqual(self.instance.vm_state, 'error')

        # Reboot the server to restart it without the share.
        self._reboot_server(server, hard=True)

        self.instance = instance.Instance.get_by_uuid(
            self.context, server['id'])
        self.assertEqual(self.instance.vm_state, 'active')

    def test_server_share_umount_failure(self):
        self.mock_disconnect.side_effect = processutils.ProcessExecutionError
        os.environ['OS_DEBUG'] = "true"
        self._check_traits()
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'
        self._attach_share(server, share_id)
        self._start_server(server)

        # Here we are using CastAsCallFixture so we got an exception from
        # nova compute. This should not happen without the fixture and
        # the api should answer with a 202 status code.
        exc = self.assertRaises(
            client.OpenStackApiException,
            self._stop_server,
            server,
        )

        self.assertIn("nova.exception.ShareUmountError", str(exc))

        log_out = self.stdlog.logger.output

        self.assertIn(
            "Share id 4b021746-d0eb-4031-92aa-23c3bec182cd umount error "
            "from server",
            log_out)

        sm = share_mapping.ShareMapping.get_by_instance_uuid_and_share_id(
            self.context, server['id'], share_id)
        self.assertEqual(sm.status, 'error')
        self.instance = instance.Instance.get_by_uuid(
            self.context, server['id'])
        self.assertEqual(self.instance.vm_state, 'error')
        return (server, share_id)

    def test_server_resume_with_shares(self):
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'
        self._attach_share(server, share_id)
        self._start_server(server)
        self.assertRaises(
            client.OpenStackApiException,
            self._suspend_server,
            server,
        )

    def test_server_rescue_with_shares(self):
        server = self._create_server(networks='auto')
        self._stop_server(server)

        share_id = '4b021746-d0eb-4031-92aa-23c3bec182cd'
        self._attach_share(server, share_id)
        self._start_server(server)
        self._rescue_server(server)

        self._assert_filesystem_tag(self._get_xml(server), share_id)

        self._assert_share_in_metadata(
            self._get_metadata_url(server), share_id, share_id)

        self._unrescue_server(server)

        self._assert_filesystem_tag(self._get_xml(server), share_id)

        self._assert_share_in_metadata(
            self._get_metadata_url(server), share_id, share_id)
        return (server, share_id)



# Copyright (c) 2013 Intel, Inc.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import glob
import os
import re
import typing as ty

from oslo_log import log as logging

from nova import exception

if ty.TYPE_CHECKING:
    # avoid circular import
    from nova.pci import stats

LOG = logging.getLogger(__name__)

PCI_VENDOR_PATTERN = "^(hex{4})$".replace("hex", r"[\da-fA-F]")
_PCI_ADDRESS_PATTERN = ("^(hex{4}):(hex{2}):(hex{2}).(oct{1})$".
                                             replace("hex", r"[\da-fA-F]").
                                             replace("oct", "[0-7]"))
_PCI_ADDRESS_REGEX = re.compile(_PCI_ADDRESS_PATTERN)
_SRIOV_TOTALVFS = "sriov_totalvfs"


def pci_device_prop_match(
    pci_dev: 'stats.Pool', specs: ty.List[ty.Dict[str, str]],
) -> bool:
    """Check if the pci_dev meet spec requirement

    Specs is a list of PCI device property requirements.
    An example of device requirement that the PCI should be either:
    a) Device with vendor_id as 0x8086 and product_id as 0x8259, or
    b) Device with vendor_id as 0x10de and product_id as 0x10d8:

    [{"vendor_id":"8086", "product_id":"8259"},
     {"vendor_id":"10de", "product_id":"10d8",
      "capabilities_network": ["rx", "tx", "tso", "gso"]}]

    """

    def _matching_devices(spec: ty.Dict[str, str]) -> bool:
        for k, v in spec.items():
            pci_dev_v = pci_dev.get(k)
            if isinstance(v, list) and isinstance(pci_dev_v, list):
                if not all(x in pci_dev.get(k) for x in v):
                    return False
            else:
                # We don't need to check case for tags in order to avoid any
                # mismatch with the tags provided by users for port
                # binding profile and the ones configured by operators
                # with pci whitelist option.
                if isinstance(v, str):
                    v = v.lower()
                if isinstance(pci_dev_v, str):
                    pci_dev_v = pci_dev_v.lower()
                if pci_dev_v != v:
                    return False
        return True

    return any(_matching_devices(spec) for spec in specs)


def parse_address(address: str) -> ty.Sequence[str]:
    """Parse a PCI address.

    Returns (domain, bus, slot, function) from PCI address that is stored in
    PciDevice DB table.
    """
    m = _PCI_ADDRESS_REGEX.match(address)
    if not m:
        raise exception.PciDeviceWrongAddressFormat(address=address)
    return m.groups()


def get_pci_address_fields(pci_addr: str) -> ty.Tuple[str, str, str, str]:
    """Parse a fully-specified PCI device address.

    Does not validate that the components are valid hex or wildcard values.

    :param pci_addr: A string of the form "<domain>:<bus>:<slot>.<function>".
    :return: A 4-tuple of strings ("<domain>", "<bus>", "<slot>", "<function>")
    """
    dbs, sep, func = pci_addr.partition('.')
    domain, bus, slot = dbs.split(':')
    return domain, bus, slot, func


def get_pci_address(domain: str, bus: str, slot: str, func: str) -> str:
    """Assembles PCI address components into a fully-specified PCI address.

    Does not validate that the components are valid hex or wildcard values.

    :param domain, bus, slot, func: Hex or wildcard strings.
    :return: A string of the form "<domain>:<bus>:<slot>.<function>".
    """
    return '%s:%s:%s.%s' % (domain, bus, slot, func)


def get_function_by_ifname(ifname: str) -> ty.Tuple[ty.Optional[str], bool]:
    """Given the device name, returns the PCI address of a device
    and returns True if the address is in a physical function.
    """
    dev_path = "/sys/class/net/%s/device" % ifname
    sriov_totalvfs = 0
    if os.path.isdir(dev_path):
        try:
            # sriov_totalvfs contains the maximum possible VFs for this PF
            with open(os.path.join(dev_path, _SRIOV_TOTALVFS)) as fd:
                sriov_totalvfs = int(fd.read())
                return (os.readlink(dev_path).strip("./"),
                        sriov_totalvfs > 0)
        except (IOError, ValueError):
            return os.readlink(dev_path).strip("./"), False
    return None, False


def is_physical_function(
    domain: str, bus: str, slot: str, function: str,
) -> bool:
    dev_path = "/sys/bus/pci/devices/%(d)s:%(b)s:%(s)s.%(f)s/" % {
        "d": domain, "b": bus, "s": slot, "f": function}
    if os.path.isdir(dev_path):
        try:
            with open(dev_path + _SRIOV_TOTALVFS) as fd:
                sriov_totalvfs = int(fd.read())
                return sriov_totalvfs > 0
        except (IOError, ValueError):
            pass
    return False


def _get_sysfs_netdev_path(pci_addr: str, pf_interface: bool) -> str:
    """Get the sysfs path based on the PCI address of the device.

    Assumes a networking device - will not check for the existence of the path.
    """
    if pf_interface:
        return "/sys/bus/pci/devices/%s/physfn/net" % pci_addr
    return "/sys/bus/pci/devices/%s/net" % pci_addr


def get_ifname_by_pci_address(
    pci_addr: str, pf_interface: bool = False,
) -> str:
    """Get the interface name based on a VF's pci address.

    The returned interface name is either the parent PF's or that of the VF
    itself based on the argument of pf_interface.
    """
    dev_path = _get_sysfs_netdev_path(pci_addr, pf_interface)
    try:
        dev_info = os.listdir(dev_path)
        return dev_info.pop()
    except Exception:
        raise exception.PciDeviceNotFoundById(id=pci_addr)


def get_mac_by_pci_address(pci_addr: str, pf_interface: bool = False) -> str:
    """Get the MAC address of the nic based on its PCI address.

    Raises PciDeviceNotFoundById in case the pci device is not a NIC
    """
    dev_path = _get_sysfs_netdev_path(pci_addr, pf_interface)
    if_name = get_ifname_by_pci_address(pci_addr, pf_interface)
    addr_file = os.path.join(dev_path, if_name, 'address')

    try:
        with open(addr_file) as f:
            mac = next(f).strip()
            return mac
    except (IOError, StopIteration) as e:
        LOG.warning("Could not find the expected sysfs file for "
                    "determining the MAC address of the PCI device "
                    "%(addr)s. May not be a NIC. Error: %(e)s",
                    {'addr': pci_addr, 'e': e})
        raise exception.PciDeviceNotFoundById(id=pci_addr)


def get_vf_num_by_pci_address(pci_addr: str) -> int:
    """Get the VF number based on a VF's pci address

    A VF is associated with an VF number, which ip link command uses to
    configure it. This number can be obtained from the PCI device filesystem.
    """
    VIRTFN_RE = re.compile(r"virtfn(\d+)")
    virtfns_path = "/sys/bus/pci/devices/%s/physfn/virtfn*" % (pci_addr)
    vf_num = None

    for vf_path in glob.iglob(virtfns_path):
        if re.search(pci_addr, os.readlink(vf_path)):
            t = VIRTFN_RE.search(vf_path)
            if t:
                vf_num = t.group(1)
                break
    else:
        raise exception.PciDeviceNotFoundById(id=pci_addr)

    return int(vf_num)


def get_vf_product_id_by_pf_addr(pci_addr: str) -> str:
    """Get the VF product ID for a given PF.

    "Product ID" or Device ID in the PCIe spec terms for a PF is
    possible to retrieve via the VF Device ID field present as of
    SR-IOV 1.0 in the "3.3.11. VF Device ID (1Ah)" section. It is
    described as a field that "contains the Device ID that should
    be presented for every VF to the SI".

    It is available as of Linux kernel 4.15, commit
    7dfca15276fc3f18411a2b2182704fa1222bcb60

    :param pci_addr: A string of the form "<domain>:<bus>:<slot>.<function>".
    :return: A string containing a product ID of a VF corresponding to the PF.
    """
    sriov_vf_device_path = f"/sys/bus/pci/devices/{pci_addr}/sriov_vf_device"
    try:
        with open(sriov_vf_device_path) as f:
            vf_product_id = f.readline().strip()
    except IOError as e:
        LOG.warning(
            "Could not find the expected sysfs file for "
            "determining the VF product ID of a PCI VF by PF"
            "with addr %(addr)s. May not be a PF. Error: %(e)s",
            {"addr": pci_addr, "e": e},
        )
        raise exception.PciDeviceNotFoundById(id=pci_addr)
    if not vf_product_id:
        raise ValueError("sriov_vf_device file does not contain"
                         " a VF product ID")
    return vf_product_id


def get_pci_ids_by_pci_addr(pci_addr: str) -> ty.Tuple[str, ...]:
    """Get the product ID and vendor ID for a given PCI device.

    :param pci_addr: A string of the form "<domain>:<bus>:<slot>.<function>".
    :return: A list containing a vendor and product ids.
    """
    id_prefix = f"/sys/bus/pci/devices/{pci_addr}"
    ids: ty.List[str] = []
    for id_name in ("vendor", "product"):
        try:
            with open(os.path.join(id_prefix, id_name)) as f:
                id_value = f.readline()
                if not id_value:
                    raise ValueError(f"{id_name} file does not contain"
                                     " a valid value")
                ids.append(id_value.strip().replace("0x", ""))
        except IOError as e:
            LOG.warning(
                "Could not find the expected sysfs file for "
                f"determining the {id_name} ID of a PCI device "
                "with addr %(addr)s. Error: %(e)s",
                {"addr": pci_addr, "e": e},
            )
            raise exception.PciDeviceNotFoundById(id=pci_addr)
    return tuple(ids)



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types


create = {
    'type': 'object',
    'properties': {
        'interfaceAttachment': {
            'type': 'object',
            'properties': {
                # NOTE: This parameter is passed to the search_opts of
                # Neutron list_network API: search_opts = {'id': net_id}
                'net_id': parameter_types.network_id,
                # NOTE: This parameter is passed to Neutron show_port API
                # as a port id.
                'port_id': parameter_types.network_port_id,
                'fixed_ips': {
                    'type': 'array', 'minItems': 1, 'maxItems': 1,
                    'items': {
                        'type': 'object',
                        'properties': {
                            'ip_address': parameter_types.ip_address
                        },
                        'required': ['ip_address'],
                        'additionalProperties': False,
                    },
                },
            },
            'additionalProperties': False,
        },
    },
    'additionalProperties': False,
}

create_v249 = copy.deepcopy(create)
create_v249['properties']['interfaceAttachment']['properties']['tag'] = parameter_types.tag  # noqa: E501

# TODO(stephenfin): Remove additionalProperties in a future API version
index_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

# TODO(stephenfin): Remove additionalProperties in a future API version
show_query = {
    'type': 'object',
    'properties': {},
    'additionalProperties': True,
}

_interface_attachment = {
    'type': 'object',
    'properties': {
        'fixed_ips': {
            'type': ['null', 'array'],
            'items': {
                'type': 'object',
                'properties': {
                    'ip_address': {
                        'type': 'string',
                        'anyOf': [
                            {'format': 'ipv4'},
                            {'format': 'ipv6'},
                        ],
                    },
                    'subnet_id': {'type': 'string', 'format': 'uuid'},
                },
                'required': ['ip_address', 'subnet_id'],
                'additionalProperties': False,
            },
        },
        'mac_addr': {'type': 'string', 'format': 'mac-address'},
        'net_id': {'type': 'string', 'format': 'uuid'},
        'port_id': {'type': 'string', 'format': 'uuid'},
        'port_state': {'type': 'string'},
    },
    'required': ['fixed_ips', 'mac_addr', 'net_id', 'port_id', 'port_state'],
    'additionalProperties': False,
}

_interface_attachment_v270 = copy.deepcopy(_interface_attachment)
_interface_attachment_v270['properties']['tag'] = {
    'type': ['null', 'string'],
}
_interface_attachment_v270['required'].append('tag')

index_response = {
    'type': 'object',
    'properties': {
        'interfaceAttachments': {
            'type': 'array',
            'items': copy.deepcopy(_interface_attachment),
        },
    },
    'required': ['interfaceAttachments'],
    'additionalProperties': False,
}

index_response_v270 = copy.deepcopy(index_response)
index_response_v270['properties']['interfaceAttachments']['items'] = copy.deepcopy(  # noqa: E501
    _interface_attachment_v270
)

show_response = {
    'type': 'object',
    'properties': {
        'interfaceAttachment': copy.deepcopy(_interface_attachment),
    },
    'required': ['interfaceAttachment'],
    'additionalProperties': False,
}

show_response_v270 = copy.deepcopy(show_response)
show_response_v270['properties']['interfaceAttachment'] = copy.deepcopy(
    _interface_attachment_v270
)

# create responses are identical to show, including microversions
create_response = copy.deepcopy(show_response)
create_response_v270 = copy.deepcopy(show_response_v270)

delete_response = {
    'type': 'null',
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock

from oslo_utils.fixture import uuidsentinel as uuids

from nova.tests.unit.virt.libvirt.volume import test_mount
from nova.tests.unit.virt.libvirt.volume import test_volume
from nova import utils
from nova.virt.libvirt.volume import cephfs
from nova.virt.libvirt.volume import mount


class LibvirtCEPHFSVolumeDriverTestCase(test_volume.LibvirtVolumeBaseTestCase):
    """Tests the libvirt CEPHFS volume driver."""

    def setUp(self):
        super(LibvirtCEPHFSVolumeDriverTestCase, self).setUp()

        self.useFixture(test_mount.MountFixture(self))

        m = mount.get_manager()
        m._reset_state()

        self.mnt_base = '/mnt'
        m.host_up(self.fake_host)
        self.flags(ceph_mount_point_base=self.mnt_base, group='libvirt')

    def test_libvirt_ceph_driver(self):
        libvirt_driver = cephfs.LibvirtCEPHFSVolumeDriver(self.fake_host)

        export_string = '192.168.1.1:/ceph/share1'
        export_mnt_base = os.path.join(self.mnt_base,
                utils.get_hash_str(export_string))

        connection_info = {
            "data": {
                "export": export_string,
                "name": self.name,
                "options": ["name=myname", "secret=mysecret"],
            }
        }
        instance = mock.sentinel.instance
        instance.uuid = uuids.instance
        libvirt_driver.connect_volume(connection_info, instance)
        libvirt_driver.disconnect_volume(connection_info,
                                         mock.sentinel.instance)

        device_path = os.path.join(export_mnt_base,
                                   connection_info['data']['name'])
        self.assertEqual(connection_info['data']['device_path'], device_path)
        self.mock_ensure_tree.assert_has_calls([mock.call(export_mnt_base)])
        self.mock_mount.assert_has_calls(
            [
                mock.call(
                    "ceph",
                    export_string,
                    export_mnt_base,
                    ["-o", "name=myname,secret=mysecret"],
                )
            ]
        )
        self.mock_umount.assert_has_calls([mock.call(export_mnt_base)])
        self.mock_rmdir.assert_has_calls([mock.call(export_mnt_base)])

    def test_libvirt_ceph_driver_with_default_options(self):
        libvirt_driver = cephfs.LibvirtCEPHFSVolumeDriver(self.fake_host)
        self.flags(
            ceph_mount_options=["default_opt1=true", "default_opt2=true"],
            group="libvirt",
        )

        export_string = '192.168.1.1:/ceph/share1'
        export_mnt_base = os.path.join(self.mnt_base,
                utils.get_hash_str(export_string))

        connection_info = {
            "data": {
                "export": export_string,
                "name": self.name,
                "options": ["name=myname", "secret=mysecret"],
            }
        }
        instance = mock.sentinel.instance
        instance.uuid = uuids.instance
        libvirt_driver.connect_volume(connection_info, instance)
        libvirt_driver.disconnect_volume(connection_info,
                                         mock.sentinel.instance)

        device_path = os.path.join(export_mnt_base,
                                   connection_info['data']['name'])
        self.assertEqual(connection_info['data']['device_path'], device_path)
        self.mock_ensure_tree.assert_has_calls([mock.call(export_mnt_base)])
        self.mock_mount.assert_has_calls(
            [
                mock.call(
                    "ceph",
                    export_string,
                    export_mnt_base,
                    [
                        "-o",
                        "default_opt1=true,default_opt2=true,name=myname,"
                        "secret=mysecret",
                    ],
                )
            ]
        )
        self.mock_umount.assert_has_calls([mock.call(export_mnt_base)])
        self.mock_rmdir.assert_has_calls([mock.call(export_mnt_base)])
